package explorviz.live_trace_processing.reader;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.TimeUnit;

import com.lmax.disruptor.RingBuffer;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.filter.RecordArrayEvent;
import explorviz.live_trace_processing.filter.counting.CountingThroughputFilter;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.normal.AfterFailedOperationEventRecord;
import explorviz.live_trace_processing.record.event.normal.AfterOperationEventRecord;
import explorviz.live_trace_processing.record.event.normal.BeforeOperationEventRecord;
import explorviz.live_trace_processing.record.misc.HostApplicationMetaDataRecord;
import explorviz.live_trace_processing.record.misc.StringRegistryRecord;
import explorviz.live_trace_processing.record.misc.SystemMonitoringRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;

public class TCPReaderOneClient extends Thread implements IPeriodicTimeSignalReceiver {
	private HostApplicationMetaDataRecord hostApplicationMetadata;

	private final Map<Integer, String> stringRegistry = new TreeMap<Integer, String>();
	private final List<byte[]> waitingForStringMessages = new ArrayList<byte[]>(1024);

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Received records/sec in Reader" + Thread.currentThread().getId());

	private final SocketChannel socketChannel;
	private final RingBuffer<RecordArrayEvent> ringBuffer;

	private IRecord[] outputBuffer = new IRecord[TCPReader.OUTPUT_MESSAGE_BUFFER_SIZE];
	private int outputBufferIndex = 0;

	public TCPReaderOneClient(final SocketChannel socketChannel,
			final RingBuffer<RecordArrayEvent> ringBuffer) {
		this.socketChannel = socketChannel;
		this.ringBuffer = ringBuffer;

		new TimeSignalReader(TimeUnit.SECONDS.toMillis(1), this).start();
	}

	@Override
	public void run() {
		final ByteBuffer buffer = ByteBuffer
				.allocateDirect(Constants.MONITORING_MESSAGE_BUFFER_SIZE);
		try {
			while ((socketChannel.read(buffer)) != -1) {
				buffer.flip();
				messagesfromByteArray(buffer);
			}
		} catch (final IOException ex) {
			System.out.println("Error in read() " + ex.getMessage());
		}
	}

	@Override
	public void periodicTimeSignal(final long timestamp) {
		// TODO flush out buffer!

		final long hiseq = ringBuffer.next();
		final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
		final IRecord[] buffer = valueEvent.getValues();
		buffer[0] = new TimedPeriodRecord();

		valueEvent.setValues(buffer);
		valueEvent.setValuesLength(1);
		valueEvent.setMetadata(hostApplicationMetadata);

		ringBuffer.publish(hiseq);
	}

	private final void messagesfromByteArray(final ByteBuffer buffer) {
		while (buffer.remaining() > 0) {
			final byte clazzId = buffer.get();
			switch (clazzId) {
				case HostApplicationMetaDataRecord.CLAZZ_ID: {
					if (buffer.remaining() >= (HostApplicationMetaDataRecord.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInTraceMetadata(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case BeforeOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= (BeforeOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInBeforeOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case AfterFailedOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= (AfterFailedOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInAfterFailedOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case AfterOperationEventRecord.CLAZZ_ID: {
					if (buffer.remaining() >= (AfterOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInAfterOperationEvent(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				case StringRegistryRecord.CLAZZ_ID: {
					int mapId = 0;
					int stringLength = 0;
					if (buffer.remaining() >= 8) {
						mapId = buffer.getInt();
						stringLength = buffer.getInt();
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}

					if (buffer.remaining() >= stringLength) {
						final byte[] stringByteArray = new byte[stringLength];

						buffer.get(stringByteArray);

						addToRegistry(mapId, new String(stringByteArray));
					} else {
						buffer.position(buffer.position() - 9);
						buffer.compact();
						return;
					}
					break;
				}
				case SystemMonitoringRecord.CLAZZ_ID: {
					if (buffer.remaining() >= (SystemMonitoringRecord.BYTE_LENGTH_WITH_CLAZZ_ID - 1)) {
						readInSystemMonitoringRecord(buffer);
					} else {
						buffer.position(buffer.position() - 1);
						buffer.compact();
						return;
					}
					break;
				}
				default: {
					System.out.println("unknown class id " + clazzId + " at offset "
							+ (buffer.position() - 4));
					buffer.clear();
					return;
				}
			}
		}

		buffer.clear();
	}

	private final void readInTraceMetadata(final ByteBuffer buffer) {
		final int hostnameId = buffer.getInt();
		final int applicationId = buffer.getInt();

		final String hostname = getStringFromRegistry(hostnameId);
		final String application = getStringFromRegistry(applicationId);

		if ((hostname != null) && (application != null)) {
			hostApplicationMetadata = new HostApplicationMetaDataRecord(hostname, application);
		} else {
			final byte[] message = new byte[HostApplicationMetaDataRecord.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- HostApplicationMetaDataRecord.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInBeforeOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);

		if (operation != null) {
			putInRingBuffer(new BeforeOperationEventRecord(timestamp, traceId, orderIndex,
					operationId, operation));
		} else {
			final byte[] message = new byte[BeforeOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- BeforeOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterFailedOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();
		final int causeId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		final String cause = getStringFromRegistry(causeId);

		if ((operation != null) && (cause != null)) {
			putInRingBuffer(new AfterFailedOperationEventRecord(timestamp, traceId, orderIndex,
					operationId, operation, cause));
		} else {
			final byte[] message = new byte[AfterFailedOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position()
					- AfterFailedOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}
	}

	private final void readInAfterOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		if (operation != null) {
			putInRingBuffer(new AfterOperationEventRecord(timestamp, traceId, orderIndex,
					operationId, operation));
		} else {
			final byte[] message = new byte[AfterOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID];
			buffer.position(buffer.position() - AfterOperationEventRecord.BYTE_LENGTH_WITH_CLAZZ_ID);
			buffer.get(message);
			putInWaitingMessages(message);
		}

	}

	private final void readInSystemMonitoringRecord(final ByteBuffer buffer) {
		final double cpuUtil = buffer.getDouble();
		final long usedRAM = buffer.getLong();
		final long absoluteRAM = buffer.getLong();
		System.out.println(new SystemMonitoringRecord(cpuUtil, usedRAM, absoluteRAM).toString());
		// putInRingBuffer(new SystemMonitoringRecord(cpuUtil, usedRAM,
		// absoluteRAM));
	}

	private final void putInWaitingMessages(final byte[] message) {
		waitingForStringMessages.add(message);
	}

	private final void checkWaitingMessages() {
		final List<byte[]> localWaitingList = new ArrayList<byte[]>();
		for (final byte[] waitingMessage : waitingForStringMessages) {
			localWaitingList.add(waitingMessage);
		}
		waitingForStringMessages.clear();

		for (final byte[] waitingMessage : localWaitingList) {
			final ByteBuffer buffer = ByteBuffer.wrap(waitingMessage);
			final byte waitingMessageClazzId = buffer.get();
			switch (waitingMessageClazzId) {
				case HostApplicationMetaDataRecord.CLAZZ_ID:
					readInTraceMetadata(buffer);
					break;
				case BeforeOperationEventRecord.CLAZZ_ID:
					readInBeforeOperationEvent(buffer);
					break;
				case AfterFailedOperationEventRecord.CLAZZ_ID:
					readInAfterFailedOperationEvent(buffer);
					break;
				case AfterOperationEventRecord.CLAZZ_ID:
					readInAfterOperationEvent(buffer);
					break;
				default:
					break;
			}
		}
	}

	private final void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);

		outputBuffer[outputBufferIndex++] = message;
		if (outputBufferIndex == TCPReader.OUTPUT_MESSAGE_BUFFER_SIZE) {
			flushOutputBuffer();
		}
	}

	private void flushOutputBuffer() {
		if (outputBufferIndex > 0) {
			final long hiseq = ringBuffer.next();
			final RecordArrayEvent valueEvent = ringBuffer.get(hiseq);
			final IRecord[] oldValues = valueEvent.getValues();

			valueEvent.setValues(outputBuffer);
			valueEvent.setValuesLength(outputBufferIndex);
			valueEvent.setMetadata(hostApplicationMetadata);
			ringBuffer.publish(hiseq);

			outputBuffer = oldValues;

			outputBufferIndex = 0;
		}
	}

	private final void addToRegistry(final int key, final String value) {
		stringRegistry.put(key, value);

		checkWaitingMessages();
	}

	private final String getStringFromRegistry(final int id) {
		return stringRegistry.get(id);
	}

	public void terminate() {
		// TODO
	}
}
